/*
 *    This file is part of LiSA Command Line Interface.
 *
 *    LiSA Command Line Interface is free software; you can redistribute it 
 *    and/or modify it under the terms of the GNU General Public License 
 *    as published by the Free Software Foundation; either version 2 of the 
 *    License, or (at your option) any later version.
 *
 *    LiSA Command Line Interface is distributed in the hope that it will be 
 *    useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with LiSA Command Line Interface; if not, write to the Free 
 *    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *    MA  02111-1307  USA
 */

#include <stdio.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <sys/param.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <curses.h>

#include "command.h"
#include "climain.h"
#include "config.h"
#include "build_config.h"
#include "filter.h"
#include "if.h"
#include "shared.h"
#include "ip.h"
#include "list.h"
#include "cdp.h"

int console_session = 0;

static char if_name[IFNAMSIZ];

static void show_ip(FILE *, char *);

static void swcli_sig_term(int sig) {
	char hostname[MAX_HOSTNAME];
	cmd_root = &command_root_main;
	signal(SIGTSTP, SIG_IGN);
	printf("^Z\n");
	gethostname(hostname, sizeof(hostname));
	hostname[sizeof(hostname) - 1] = '\0';
	sprintf(prompt, cmd_root->prompt, hostname, priv ? '#' : '>');
	rl_set_prompt(prompt);
	rl_forced_update_display();
}

/* Command Handlers implementation */
static void cmd_disable(FILE *out, char **argv) {
	int req;
	
	req = argv[0] == NULL ? 1 : atoi(argv[0]);
	if (req > priv) {
		fprintf(out, 
				"New privilege level must be less than current"
				"privilege level\n");
		return;
	}
	priv = req;
}

static int secret_validator(char *pass, void *arg) {
	char *secret = arg;

	return !strcmp(secret, crypt(pass, secret));
}

static void cmd_enable(FILE *out, char **argv) {
	int req;
	int fail = 0;
	char secret[CLI_SECRET_LEN + 1];

	req = argv[0] == NULL ? 15 : atoi(argv[0]);
	if(req > priv) {
		fail = 1;
		cfg_lock();
		strcpy(secret, cfg->enable_secret[req]);
		cfg_unlock();
		if(secret[0] == '\0') {
			if(console_session) {
				priv = req;
				return;
			}
			fprintf(out, "%% No password set\n");
			return;
		}
		fail = !cfg_checkpass(3, secret_validator, secret);
	}
	if(fail) {
		fprintf(out, "%% Bad secrets\n\n");
		return;
	}
	priv = req;
}

static void cmd_conf_t(FILE *out, char **argv) {
	cmd_root = &command_root_config;
	printf("Enter configuration commands, one per line.  End with CNTL/Z.\n");
	signal(SIGTSTP, swcli_sig_term);
}

void cmd_help(FILE *out, char **argv) {
	fprintf(out,
		"Help may be requested at any point in a command by entering\n"
		"a question mark '?'.  If nothing matches, the help list will\n"
		"be empty and you must backup until entering a '?' shows the\n"
		"available options.\n"
		"Two styles of help are provided:\n"
		"1. Full help is available when you are ready to enter a\n"
		"   command argument (e.g. 'show ?') and describes each possible\n"
		"   argument.\n"
		"2. Partial help is provided when an abbreviated argument is entered\n"
		"   and you want to know what arguments match the input\n"
		"   (e.g. 'show pr?'.)\n\n"
		);
}

void cmd_show_ver(FILE *out, char **argv) {
	FILE *fp;
	char buf[255];
	
	fprintf(out,
		"LiSA Command Line Interface, ver. 1.0, compiled 2005.06.01 \n"
		"For more info go to http://lisa.ines.ro\n\n"
		"This tool is part of the Linux Switch Appliance Project and can\n"
		"be used (succesfully) only with a kernel patched with the LMS\n"
		"(i.e. Linux Multilayer Switch) kernel patch.\n\n"
		"Operating system info:\n\n"
			);
	fp = fopen(VERSION_FILE_PATH, "r");
	if (!fp) {
		fprintf(out, "No available information.\n\n");
		return;
	}
	while (fgets(buf, sizeof(buf), fp)) {
		fprintf(out, "%s\n", buf);
	}
	fclose(fp);
	fprintf(out, "\n");
}

static void cmd_history(FILE *out, char **argv) {
	HIST_ENTRY **history;
	HIST_ENTRY *entry;
	int i;

	history = history_list();
	if (history) {
		for (i=0; (entry = history[i]); i++) {
			fprintf(out, "   %s\n", entry->line);
		}
	}
}

static void cmd_exit(FILE *out, char **argv) {
	pclose(out);
	exit(0);
}

static void cmd_run_eth(FILE *out, char **argv) {
	char *arg = *argv;
	int status;
	FILE *tmp = NULL;
	char tmp_name[MAXPATHLEN];

	fprintf(out, "Building configuration...\n");
	fflush(out);
	do {
		tmp = mk_tmp_stream(tmp_name, "w+");
		if(tmp == NULL)
			break;
		status = build_int_eth_config(tmp, parse_eth(arg));
		if(status)
			break;
		fprintf(out, "\nCurrent configuration : %ld bytes\n", ftell(tmp));
		rewind(tmp);
		copy_data(out, tmp);
		fclose(tmp);
		unlink(tmp_name);
		fprintf(out, "end\n\n");
		fflush(out);
		return;
	} while(0);
	if(tmp != NULL) {
		fclose(tmp);
		unlink(tmp_name);
	}
	fprintf(out, "Command rejected: device not present\n");
	return;
}

static void cmd_show_priv(FILE *out, char **argv) {
	fprintf(out, "Current privilege level is %d\n", priv);
	fflush(out);
}

static void cmd_show_start(FILE *out, char **argv) {
	char cfg_name[] = "/flash/config.text";
	char buf[1024];
	FILE *fp;

	fp = fopen(cfg_name, "r");
	while (fgets(buf, sizeof(buf), fp)) {
		buf[sizeof(buf)-1] = '\0';
		fprintf(out, buf);
	}
	fclose(fp);
}

static void cmd_show_run(FILE *out, char **argv) {
	int status;
	FILE *tmp = NULL;
	char tmp_name[MAXPATHLEN];

	fprintf(out, "Building configuration...\n");
	fflush(out);
	do {
		tmp = mk_tmp_stream(tmp_name, "w+");
		if(tmp == NULL)
			break;
		status = build_config(tmp);
		if(status)
			break;
		fprintf(out, "\nCurrent configuration : %ld bytes\n", ftell(tmp));
		rewind(tmp);
		copy_data(out, tmp);
		fclose(tmp);
		unlink(tmp_name);
		fprintf(out, "!\nend\n\n");
		fflush(out);
		return;
	} while(0);
	if(tmp != NULL) {
		fclose(tmp);
		unlink(tmp_name);
	}
	fprintf(out, "Oops! Something went terribly wrong :(\n");
	return;
}

static void cmd_run_vlan(FILE *out, char **argv) {
	char *dev = NULL;
	if (*argv)
		dev = if_name_vlan(*argv);
	show_ip(out, dev);
}

static void init_mac_filter(struct net_switch_ioctl_arg *user_arg) {
	user_arg->if_name = NULL;
	user_arg->cmd = SWCFG_GETMAC;
	memset(&user_arg->ext.marg.addr, 0, ETH_ALEN);
	user_arg->ext.marg.addr_type = SW_FDB_ANY;
	user_arg->vlan = 0;
}

static void do_mac_filter(FILE *out, struct net_switch_ioctl_arg *user_arg, int size, char *buf) {
	int status;

	do {
		user_arg->ext.marg.buf_size = size;
		user_arg->ext.marg.buf = buf;
		status = ioctl(sock_fd, SIOCSWCFG, user_arg);
		if (status == -1) {
			if (errno == ENOMEM) {
				dbg("Insufficient buffer space. Realloc'ing ...\n");
				buf = realloc(buf, size+INITIAL_BUF_SIZE);
				assert(buf);
				size+=INITIAL_BUF_SIZE;
				continue;
			}
			perror("ioctl");
			return;
		}
	} while(status < 0);
	user_arg->ext.marg.actual_size = status;
	cmd_showmac(out, (char *)user_arg);
}

static void cmd_show_mac(FILE *out, char **argv) {
	int i, size;
	char *arg, *buf;
	struct net_switch_ioctl_arg user_arg;

	buf = (char *)malloc(INITIAL_BUF_SIZE);
	size = INITIAL_BUF_SIZE;
	init_mac_filter(&user_arg);
	for (i=0; (arg = argv[i]); i++) {
		if (strcmp(arg, "dynamic") == 0) {
			user_arg.ext.marg.addr_type = SW_FDB_DYN;
			continue;
		}
		if (strcmp(arg, "static") == 0) {
			user_arg.ext.marg.addr_type = SW_FDB_STATIC;
		}
	}
	do_mac_filter(out, &user_arg, size, buf);
	fprintf(out, "\n");
}

static void cmd_sh_mac_eth(FILE *out, char **argv) {
	char *arg, *buf;
	struct net_switch_ioctl_arg user_arg;
	int i, size;


	buf = (char *)malloc(INITIAL_BUF_SIZE);
	size = INITIAL_BUF_SIZE;
	init_mac_filter(&user_arg);

	for (i=0; (arg = argv[i]); i++) {
		if (strcmp(arg, "dynamic") == 0) {
			user_arg.ext.marg.addr_type = SW_FDB_DYN;
			continue;
		}
		if (strcmp(arg, "static") == 0) {
			user_arg.ext.marg.addr_type = SW_FDB_STATIC;
			continue;
		}
		if (valid_eth(arg, arg[strlen(arg)-1]) && !user_arg.if_name) {
			user_arg.if_name = if_name_eth(arg);
		}
	}

	do_mac_filter(out, &user_arg, size, buf);
	fprintf(out, "\n");
}

static void cmd_sh_mac_vlan(FILE *out, char **argv) {
	char *arg, *buf;
	struct net_switch_ioctl_arg user_arg;
	int i, size, a[2], cnt=0, k;

	
	buf = (char *)malloc(INITIAL_BUF_SIZE);
	size = INITIAL_BUF_SIZE;
	init_mac_filter(&user_arg);
	
	for (i=0; (arg = argv[i]); i++) {
		if (strcmp(arg, "dynamic") == 0) {
			user_arg.ext.marg.addr_type = SW_FDB_DYN;
			continue;
		}
		if (strcmp(arg, "static") == 0) {
			user_arg.ext.marg.addr_type = SW_FDB_STATIC;
			continue;
		}
		if (sscanf(arg, "%d", &k)) {
			a[cnt++] = k;
		}
	}
	if (cnt > 1) { /* si eth si vlan */
		sprintf(if_name, "eth%d", a[0]);
		user_arg.if_name = if_name;
		user_arg.vlan = a[1];
	}
	else user_arg.vlan = a[0];

	do_mac_filter(out, &user_arg, size, buf);
	fprintf(out, "\n");
}

static void cmd_sh_addr(FILE *out, char **argv) {
	char *arg, *buf;
	struct net_switch_ioctl_arg user_arg;
	int i, size;
	
	buf = (char *)malloc(INITIAL_BUF_SIZE);
	size = INITIAL_BUF_SIZE;
	init_mac_filter(&user_arg);
	
	for (i=0; (arg = argv[i]); i++) {
		if (strcmp(arg, "dynamic") == 0) {
			user_arg.ext.marg.addr_type = SW_FDB_DYN;
			continue;
		}
		if (strcmp(arg, "static") == 0) {
			user_arg.ext.marg.addr_type = SW_FDB_STATIC;
			continue;
		}
		if (valid_mac(arg, ' ')) {
			parse_mac(arg, user_arg.ext.marg.addr);
		}
	}
	
	do_mac_filter(out, &user_arg, size, buf);
	fprintf(out, "\n");
}

static void cmd_ping(FILE *out, char **argv) {
	FILE *p;
	char cmd_buf[MAX_LINE_WIDTH];
	int nc;

	memset(cmd_buf, 0, sizeof(cmd_buf));
	nc = sprintf(cmd_buf, "%s -i %d -c %d %s", PING_PATH, PING_INTERVAL, PING_COUNT, argv[0]);
	assert(nc < sizeof(cmd_buf));
	assert((p =	popen(cmd_buf, "w")));
	pclose(p);
}

static void cmd_trace(FILE *out, char **argv) {
	FILE *p;
	char cmd_buf[MAX_LINE_WIDTH];
	int nc;

	memset(cmd_buf, 0, sizeof(cmd_buf));
	nc = sprintf(cmd_buf, "%s %s", TRACEROUTE_PATH, argv[0]);
	assert(nc < sizeof(cmd_buf));
	assert((p =	popen(cmd_buf, "w")));
	pclose(p);
}

static void cmd_clr_mac_eth(FILE *out, char **argv) {
	struct net_switch_ioctl_arg ioctl_arg;
	int status;
	unsigned char mac[ETH_ALEN];

	ioctl_arg.cmd = SWCFG_DELMACDYN;
	ioctl_arg.vlan = 0;
	memset(mac, 0, ETH_ALEN);
	ioctl_arg.ext.mac = mac;
	ioctl_arg.if_name = if_name_eth(argv[0]);
	status = ioctl(sock_fd, SIOCSWCFG, &ioctl_arg);
	if (status == -1) {
		fprintf(out, "MAC address could not be removed\n"
				"Address not found\n\n");
		fflush(out);
	}
}

static void cmd_clr_mac(FILE *out, char **argv) {
	struct net_switch_ioctl_arg ioctl_arg;
	int status;
	unsigned char mac[ETH_ALEN];

	ioctl_arg.cmd = SWCFG_DELMACDYN;
	memset(mac, 0, ETH_ALEN);
	ioctl_arg.ext.mac = mac;
	ioctl_arg.vlan = 0;
	ioctl_arg.if_name = NULL;
	status = ioctl(sock_fd, SIOCSWCFG, &ioctl_arg);
	if (status == -1) {
		fprintf(out, "MAC address could not be removed\n"
				"Address not found\n\n");
		fflush(out);
	}
}

static void cmd_clr_mac_vl(FILE *out, char **argv) {
	struct net_switch_ioctl_arg ioctl_arg;
	int status;
	unsigned char mac[ETH_ALEN];

	ioctl_arg.cmd = SWCFG_DELMACDYN;
	ioctl_arg.vlan = parse_vlan(argv[0]);
	memset(mac, 0, ETH_ALEN);
	ioctl_arg.ext.mac = mac;
	ioctl_arg.if_name = NULL;
	status = ioctl(sock_fd, SIOCSWCFG, &ioctl_arg);
	if (status == -1) {
		fprintf(out, "MAC address could not be removed\n"
				"Address not found\n\n");
		fflush(out);
	}
}

static void cmd_clr_mac_adr(FILE *out, char **argv) {
	struct net_switch_ioctl_arg ioctl_arg;
	int status;
	unsigned char mac[ETH_ALEN];

	ioctl_arg.cmd = SWCFG_DELMACDYN;
	status = parse_mac(argv[0], mac);
	assert(!status);
	ioctl_arg.ext.mac = mac;
	ioctl_arg.vlan = 0;
	ioctl_arg.if_name = NULL;
	status = ioctl(sock_fd, SIOCSWCFG, &ioctl_arg);
	if (status == -1) {
		fprintf(out, "MAC address could not be removed\n"
				"Address not found\n\n");
		fflush(out);
	}
}

static void show_ip_list(FILE *out, struct list_head *ipl) {
	struct ip_addr_entry *entry, *tmp;

	list_for_each_entry_safe(entry, tmp, ipl, lh) {
		fprintf(out, "  inet: %s / %s\n", 
				entry->inet, entry->mask);
		list_del(&entry->lh);
		free(entry);
	}
}

static void show_ip(FILE *out, char *dev) {
	FILE *fh;
	char buf[128];
	struct list_head *ipl;

	fh = fopen(PROCNETSWITCH_PATH, "r");
	if (!fh) {
		perror("fopen");
		exit(-1);
	}
	while (fgets(buf, sizeof(buf), fh)) {
		/* make sure buf is null-terminated */
		buf[sizeof(buf)-1] = '\0';
		/* strip the newline at the end of buf */
		if (buf && strlen(buf))
			buf[strlen(buf)-1] = '\0'; 
		/* compare to the interface name we're searching for */
		if (dev && strcmp(dev, buf))
			continue;
		ipl = list_ip_addr(buf, 0);
		if (ipl && !list_empty(ipl)) {
			fprintf(out, "%s:\n", buf);
			show_ip_list(out, ipl);
		}
		if (ipl)
			free(ipl);
		if (dev)
			break;	
	}
	fclose(fh);
}

static void cmd_sh_ip(FILE *out, char **argv) {
	char *dev = NULL;
	if (*argv)
		dev = if_name_vlan(*argv);
	show_ip(out, dev);
}

/* FIXME: quick hack
   ar tb facut ca la show mac (selectori id si name) 
 */
static void cmd_show_vlan(FILE *out, char **argv) {
	FILE *in;
	char buf[512];

	fprintf(out, "\n");
	if ((in = fopen(VLAN_FILE_PATH, "r")) == NULL) {
		perror("fopen");
		exit(-1);
	}
	
	while (fgets(buf, sizeof(buf), in)) {
		buf[strlen(buf)-1] = '\0';
		fprintf(out, "%s\n", buf);
	}

	fclose(in);
}

int get_mac_age() {
	struct net_switch_ioctl_arg ioctl_arg;
	int status;

	ioctl_arg.cmd = SWCFG_GETAGETIME;
	status = ioctl(sock_fd, SIOCSWCFG, &ioctl_arg);
	if (status) { 
		perror("Error getting age time");
		return -1;
	}	
	return ioctl_arg.ext.nsec;
}

static void cmd_sh_mac_age(FILE *out, char **argv) {
	long age;

	if ((age = get_mac_age()) >= 0) {
		fprintf(out, "%li\n", age);
	}
}

static void cmd_wrme(FILE *out, char **argv) {
	int status;
	FILE *tmp = NULL;
	char tmp_name[] = "/flash/config.text";

	fprintf(out, "Building configuration...\n");
	fflush(out);
	cfg_lock();
	do {
		tmp = fopen(tmp_name, "w+");
		if(tmp == NULL)
			break;
		status = build_config(tmp);
		if(status) {
			fclose(tmp);
			break;
		}
		fprintf(out, "Current configuration : %ld bytes\n", ftell(tmp));
		fclose(tmp);
		sync();
		fprintf(out, "\n[OK]\n\n");
	} while(0);
	cfg_unlock();
}

static void cmd_reload(FILE *out, char **argv) {
	int key;

	fputs("Proceed with reload? [confirm]", out);
	fflush(out);
	key = read_key();
	if(key != 'y' && key != 'Y' && key != '\n') {
		fputc('\n', out);
		return;
	}
	system("reboot &> /dev/null");
	/* delay returning so that the prompt does'n get displayed again */
	sleep(5);
}


/* Validation Handlers Implementation */
int valid_regex(char *arg, char lookahead) {
	return 1;
}

int valid_eth(char *arg, char lookahead) {
	int no;
	if(sscanf(arg, "%d", &no) != 1)
		return 0;
	if(no < 0 || no > 7) /* FIXME max value */
		return 0;
	/* FIXME interfata e in switch */
	return 1;
}

int valid_vlan(char *arg, char lookahead) {
	int no;
	if(sscanf(arg, "%d", &no) != 1)
		return 0;
	if(no < 1 || no > 4094)
		return 0;
	return 1;
}

int valid_mac(char *arg, char lookahead) {
	regex_t regex;
	regmatch_t result;
	int res;

	if (regcomp(&regex, whitespace(lookahead)? 
				"^[0-9a-f]{1,4}(\\.[0-9a-f]{1,4}){2}$":"^[0-9a-f]{1,4}(\\.[0-9a-f]{0,4}){0,2}$", 
				REG_ICASE|REG_EXTENDED)) {
		perror("regcomp");
		return 0;
	}
	res = regexec(&regex, arg, 1, &result, 0);
	return (res == 0);
}

int valid_dyn(char *arg, char lookahead) {
	return (strcmp(arg, "dynamic") == 0);
}

int valid_static(char *arg, char lookahead) {
	return (strcmp(arg, "static") == 0);
}

int valid_star(char *arg, char lookahead) {
	return (strcmp(arg, "*") == 0);
}

int valid_protocol(char *arg, char lookahead) {
	return (arg && strlen(arg) && strncmp(arg, "protocol", strlen(arg)) == 0);
}

int valid_version(char *arg, char lookahead) {
	return (arg && strlen(arg) && strncmp(arg, "version", strlen(arg)) == 0);
}

char vlan_range[] = "<1-1094>\0";

int parse_eth(char *arg) {
	int no, status;

	status = sscanf(arg, "%d", &no);
	assert(status == 1);

	return no;
}

char *if_name_eth(char *arg) {
	sprintf(if_name, "eth%d", parse_eth(arg));
	return if_name;
}

char *if_name_vlan(char *arg) {
	sprintf(if_name, "vlan%d", parse_vlan(arg));
	return if_name;
}

int parse_vlan(char *arg) {
	int no, status;

	status = sscanf(arg, "%d", &no);
	assert(status == 1);
	assert(no >= 1 && no <= 4094);
	return no;
}

static char vlan_name[16];
char *default_vlan_name(int vlan) {
	snprintf(vlan_name, sizeof(vlan_name), "VLAN%04d", vlan);
	return vlan_name;
}

int parse_mac(char *arg, unsigned char *mac) {
	unsigned short a0, a1, a2;

	if (sscanf(arg, "%hx.%hx.%hx", &a0, &a1, &a2) != ETH_ALEN/2)
		return EINVAL;

	mac[0] = a0 / 0x100;
	mac[1] = a0 % 0x100;
	mac[2] = a1 / 0x100;
	mac[3] = a1 % 0x100;
	mac[4] = a2 / 0x100;
	mac[5] = a2 % 0x100;

	return 0;
}

int valid_priv(char *arg, char lookahead) {
	int n;

	n = atoi(arg);
	if(n >= 1 && n <= 15)
		return 1;
	return 0;
}

static sw_command_t sh_pipe_regex[] = {
	{"LINE",				1,  valid_regex,	NULL,			RUN,		"Regular Expression",								NULL},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_pipe_mod[] = {
	{"begin",				1,  NULL,			NULL,			MODE_BEGIN,		"Begin with the line that matches",				sh_pipe_regex},
	{"exclude",				1,  NULL,			NULL,			MODE_EXCLUDE,	"Exclude lines that match",						sh_pipe_regex},
	{"include",				1,  NULL,			NULL,			MODE_INCLUDE,	"Include lines that match",						sh_pipe_regex},
	{"grep",				1,  NULL,			NULL,			MODE_GREP,		"Linux grep functionality",						sh_pipe_regex},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

sw_command_t sh_pipe[] = {
	{"|",					1,  NULL,			NULL,			0,			"Output modifiers",									sh_pipe_mod},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_int_eth[] = {
	{eth_range,				1,	valid_eth,		cmd_int_eth,	RUN,		"Ethernet interface number",						NULL},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_int_vlan[] = {
	{vlan_range,			1,	valid_vlan,		cmd_int_vlan,	RUN,		"Vlan interface number",							NULL},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_run_eth[] = {
	{eth_range,				2,	valid_eth,		cmd_run_eth,	RUN,		"Ethernet interface number",						NULL},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_run_vlan[] = {
	{vlan_range,			2,	valid_vlan,		cmd_run_vlan,	RUN,		"Vlan interface number",							NULL},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_mac_vlan[] = {
	{vlan_range,			1,	valid_vlan,		cmd_sh_mac_vlan,RUN|PTCNT,	"Vlan interface number",							sh_pipe},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_sh_m_eth[] = {
	{"vlan",				1,	NULL,			NULL,			0,			 "VLAN keyword",									sh_mac_vlan},
	{"|",					1,	NULL,			NULL,			0,			"Output modifiers",									sh_pipe_mod},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_mac_eth[] = {
	{eth_range,				0,	valid_eth,		cmd_sh_mac_eth,	RUN|PTCNT,	"Ethernet interface number",						sh_sh_m_eth},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_sh_int[] = {
	{"ethernet",			1,	NULL,			NULL,			0,			"Ethernet IEEE 802.3",								sh_int_eth},
	{"vlan",				1,	NULL,			NULL,			0,			"LMS Vlans",										sh_int_vlan},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_cdp_int_eth[] = {
	{eth_range,				1,	valid_eth,  cmd_sh_cdp_int,   RUN,			"Ethernet interface number",						NULL},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_cdp_ne_det[] = {
	{ "detail",				1,  NULL,	cmd_sh_cdp_ne_detail,	  RUN,			"Show detailed information",					sh_pipe},
	{NULL,					0,	NULL,					NULL,		0,			NULL,											NULL}
};

static sw_command_t sh_cdp_ne_eth[] = {
	{eth_range,				1,	valid_eth,  cmd_sh_cdp_ne_int, RUN|PTCNT,	"Ethernet interface number",						sh_cdp_ne_det},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_cdp_int[] = {
	{"ethernet",			1,	NULL,			NULL,			0,			"Ethernet IEEE 802.3",								sh_cdp_int_eth},
	{ "|",					1, 	NULL,			NULL,	 	    0,			"Output modifiers",									sh_pipe_mod},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_cdp_entry_ve[] = {
	{"protocol",			1, 	valid_protocol,		cmd_sh_cdp_entry,	RUN|PTCNT|CMPL,		"Protocol information",				sh_pipe},
	{ "|",					1, 	NULL,							NULL,	  			 0,		"Output modifiers",					sh_pipe_mod},
	{NULL,					0,	NULL,							NULL,	  			 0,		NULL,								NULL}
};

static sw_command_t sh_sh_cdp_entry[] = {
	{"protocol",			1, 	valid_protocol,		cmd_sh_cdp_entry,	RUN|PTCNT|CMPL,		"Protocol information",				sh_pipe},
	{"version",				1,	valid_version,		cmd_sh_cdp_entry,	RUN|PTCNT|CMPL,		"Version information",				sh_cdp_entry_ve},
	{ "|",					1,			 NULL,					NULL,	 			 0,		"Output modifiers",					sh_pipe_mod},
	{NULL,					0,			 NULL,					NULL,	 			 0,		NULL,										NULL}
};

static sw_command_t sh_cdp_entry[] = {
	{"*",					1,	valid_star,			cmd_sh_cdp_entry,		RUN|PTCNT|CMPL,		"all CDP neighbor entries",			sh_sh_cdp_entry},
	{"WORD",				1, valid_regex,			cmd_sh_cdp_entry,			 RUN|PTCNT,		"Name of CDP neighbor entry",		sh_sh_cdp_entry},
	{NULL,					0,	 NULL,							NULL,					 0,		NULL,								NULL}
};

static sw_command_t sh_cdp_ne[] = {
	{"ethernet",			1,	NULL,			NULL,				0,			"Ethernet IEEE 802.3",							sh_cdp_ne_eth},
	{ "detail",				1,  NULL,	cmd_sh_cdp_ne_detail,	  RUN,			"Show detailed information",					sh_pipe},
	{"|",					1,	NULL,			NULL,			  	0,			"Output modifiers",								sh_pipe_mod},
	{NULL,					0,	NULL,			NULL,				0,			NULL,											NULL}
};

static sw_command_t sh_cdp[] = {
	{"entry",				1,  NULL,				   NULL,			0,			"Information for specific neighbor entry",	sh_cdp_entry},
	{"holdtime",			1,	NULL,	cmd_sh_cdp_holdtime,		  RUN,			"Time CDP info kept by neighbors",			sh_pipe},
	{"interface",			1,	NULL,	cmd_sh_cdp_int,			 	  RUN,			"CDP interface status and configuration",	sh_cdp_int},
	{"neighbors",			1, 	NULL,	cmd_sh_cdp_ne,				  RUN,			"CDP neighbor entries",						sh_cdp_ne},
	{"run",					1,	NULL,	cmd_sh_cdp_run,				  RUN,			"CDP process running",						sh_pipe},
	{"timer",				1,	NULL,	cmd_sh_cdp_timer,			  RUN,			"Time CDP info is resent to neighbors",		sh_pipe},
	{"traffic",				1,	NULL,	cmd_sh_cdp_traffic,			  RUN,			"CDP statistics",							sh_pipe},
	{"|",					1, 	NULL,			NULL,					0,			"Output modifiers",							sh_pipe_mod},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_sh_run_int[] = {
	{"ethernet",			2,	NULL,			NULL,			0,			"Ethernet IEEE 802.3",								sh_run_eth},
	{"vlan",				2,	NULL,			NULL,			0,			"LMS Vlans",										sh_run_vlan},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_sh_mac_int[] = {
	{"ethernet",			1,	NULL,			NULL,			0,			"Ethernet IEEE 802.3",								sh_mac_eth},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_sh_mac_addr[] = {
	{"H.H.H",				1,	valid_mac,		cmd_sh_addr,	RUN|PTCNT,	"48 bit mac address",								NULL},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_sh_mac_sel[] = {
	{"address",				1,	NULL,			NULL,			0,			"address keyword",									sh_sh_mac_addr},
	{"interface",			1,	NULL,			NULL,			0,			"interface keyword",								sh_sh_mac_int},
	{"vlan",				1,	NULL,			NULL,			0,			 "VLAN keyword",									sh_mac_vlan},
	{"|",					1,	NULL,			NULL,			0,			"Output modifiers",									sh_pipe_mod},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_mac_addr_table[] = {
	{"address",				1,	NULL,			NULL,			0,			"address keyword",									sh_sh_mac_addr},
	{"aging-time",			1,	NULL,			cmd_sh_mac_age,			RUN,		"aging-time keyword",						sh_pipe},
	{"dynamic",				1,	valid_dyn,		cmd_show_mac,	RUN|CMPL|PTCNT,	"dynamic entry type",								sh_sh_mac_sel},
	{"interface",			1,	NULL,			NULL,			0,			"interface keyword",								sh_sh_mac_int},
	{"static",				1,	valid_static,	cmd_show_mac,	RUN|CMPL|PTCNT,	"static entry type",								sh_sh_mac_sel},
	{"vlan",				1,	NULL,			NULL,			0,			 "VLAN keyword",									sh_mac_vlan},
	{"|",					1,	NULL,			NULL,			0,			"Output modifiers",									sh_pipe_mod},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_show_run[] = {
	{"interface",			1,	NULL,			NULL,			0,			"Show interface configuration",						sh_sh_run_int},
	{"|",					1,  NULL,			NULL,			0,			"Output modifiers",									sh_pipe_mod},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_sh_mac[] = {
	{"address-table",		1,	NULL,			cmd_show_mac,	RUN,		"MAC forwarding table",								sh_mac_addr_table},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_show_vlan[] = {
	{"brief",				1, 	NULL,			cmd_show_vlan,	RUN,		"VTP all VLAN status in brief",						sh_pipe},
	{"|",					1,  NULL,			NULL,			0,			"Output modifiers",									sh_pipe_mod},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_show[] = {
	{"arp",					1,	NULL,			NULL,			RUN,		"ARP table",										NULL},
	{"clock",				1,	NULL,			NULL,			RUN,		"Display the system clock",							NULL},
	{"cdp",					1,	NULL,			cmd_sh_cdp,		RUN,		"CDP Information",									sh_cdp},
	{"history",				1,	NULL,			cmd_history,	RUN,		"Display the session command history",				sh_pipe},
	{"interfaces",			1,	NULL,			cmd_sh_int,		RUN,		"Interface status and configuration",				sh_sh_int},
	{"ip",					1,	NULL,			cmd_sh_ip,		RUN,		"IP information",									NULL},
	{"mac",					1,	NULL,			NULL,			0,			"MAC configuration",								sh_sh_mac},
	{"mac-address-table",	1,	NULL,			cmd_show_mac,	RUN,		"MAC forwarding table",								sh_mac_addr_table},
	{"privilege",			2,	NULL,			cmd_show_priv,	RUN,		"Show current privilege level",						NULL},
	{"running-config",		2,	NULL,			cmd_show_run,	RUN,		"Current operating configuration",					sh_show_run},
	{"sessions",			1,	NULL,			NULL,			RUN,		"Information about Telnet connections",				NULL},
	{"startup-config",		1,	NULL,			cmd_show_start,	RUN,		"Contents of startup configuration",				NULL},
	{"users",				1,	NULL,			NULL,			RUN,		"Display information about terminal lines",			NULL},
	{"version",				1,	NULL,			cmd_show_ver,			RUN,		"System hardware and software status",				NULL},
	{"vlan",				1,	NULL,			cmd_show_vlan,	RUN,		"VTP VLAN status",									sh_show_vlan},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_conf[] = {
	{"terminal",			2,	NULL,			cmd_conf_t,		RUN,		"Configure from the terminal",						NULL},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_ping[] = {
	{"WORD",				1,	valid_regex,	cmd_ping,		RUN|PTCNT,	"Ping destination address or hostname",				NULL},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_trace[] = {
	{"WORD",				1,	valid_regex,	cmd_trace,		RUN|PTCNT,	"Ping destination address or hostname",				NULL},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

char priv_range[] = "<1-15>\0";

static sw_command_t sh_enable[] = {
	{priv_range,			1,	valid_priv,		cmd_enable,		RUN|PTCNT|NPG,"Enable level",									NULL},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_disable[] = {
	{priv_range,			1,	valid_priv,		cmd_disable,	RUN|PTCNT|NPG,"Privilege level to go to",								NULL},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_clr_eth[] = {
	{eth_range,				2,	valid_eth,		cmd_clr_mac_eth,RUN|PTCNT,	"Ethernet interface number",						NULL},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_clr_sel_addr[] = {
	{"H.H.H",				2,	valid_mac,		cmd_clr_mac_adr,RUN|PTCNT,	"48 bit mac address",								NULL},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_clr_sel_int[] = {
	{"ethernet",			2,	NULL,			NULL,			0,			"Ethernet IEEE 802.3",								sh_clr_eth},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_clr_sel_vlan[] = {
	{vlan_range,			2,	valid_vlan,		cmd_clr_mac_vl,RUN|PTCNT,	"Vlan interface number",							NULL},
	{NULL,					0,	NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_clr_sel[] = {
	{"address",				2,	NULL,			NULL,			0,			"address keyword",									sh_clr_sel_addr},
	{"interface",			2,	NULL,			NULL,			0,			"interface keyword",								sh_clr_sel_int},
	{"vlan",				2,	NULL,			NULL,			0,			"vlan keyword",										sh_clr_sel_vlan},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_clr_mac_addr[] = {
	{"dynamic",				2,	NULL,			cmd_clr_mac,	RUN,		"dynamic entry type",								sh_clr_sel},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_clr_mac[] = {
	{"address-table",		2,	NULL,			NULL,			0,			"MAC forwarding table",								sh_clr_mac_addr},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_clear[] = {
	{"mac",					2,	NULL,			NULL,			0,			"MAC forwarding table",								sh_clr_mac},
	{"mac-address-table",	2,	NULL,			NULL,			0,			"MAC forwarding table",								sh_clr_mac_addr},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh_write[] = {
	{"memory",				15,	NULL,			cmd_wrme,		RUN,		"Write to NV memory",								NULL},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

static sw_command_t sh[] = {
	{"clear",				2,	NULL, 			NULL,			0,			"Reset functions",									sh_clear},
	{"configure",			15,	NULL,			NULL,			0,			"Enter configuration mode",							sh_conf},
	{"disable",				2,	NULL,			cmd_disable,	RUN,		"Turn off privileged commands",						sh_disable},
	{"enable",				1,	NULL,			cmd_enable,		RUN|NPG,	"Turn on privileged commands",						sh_enable},
	{"exit",				1,	NULL,			cmd_exit,		RUN,		"Exit from the EXEC",								NULL},
	{"help",				1,	NULL,			cmd_help,		RUN,		"Description of the interactive help system",		NULL},
	{"logout",				1,	NULL,			cmd_exit,		RUN,		"Exit from the EXEC",								NULL},
	{"ping",				1,	NULL,			NULL,			0,			"Send echo messages",								sh_ping},
	{"reload",				1,	NULL,			cmd_reload,		RUN|NPG,	"Halt and perform a cold restart",					NULL},
	{"quit",				1,	NULL,			cmd_exit,		RUN,		"Exit from the EXEC",								NULL},
	{"show",				1,	NULL,			NULL,			0,			"Show running system information",					sh_show},
//	{"telnet",				1,	NULL,			NULL,			0,			"Open a telnet connection",							NULL},
//	{"terminal",			1,	NULL,			NULL,			0,			"Set terminal line parameters",						NULL},
	{"traceroute",			1,	NULL,			NULL,			0,			"Trace route to destination",						sh_trace},
	{"where",				1,	NULL,			NULL,			RUN,		"List active connections",							NULL},
	{"write",				15,	NULL,			NULL,			0,			"Write running configuration to memory",			sh_write},
	{NULL,					0,  NULL,			NULL,			0,			NULL,												NULL}
};

sw_command_root_t command_root_main =					{"%s%c",					sh};
